-- |
-- Module      : Term
-- Copyright   : (c) 2009 Bertram Felgenhauer
-- License     : MIT
--
-- Maintainer  : Bertram Felgenhauer <int-e@gmx.de>
-- Stability   : experimental
-- Portability : portable
--
-- Dieses Modul stellt einen Datentyp 'Term' fuer einfache Ausdruecke
-- zur Verfuegung.
--
module Term (
    -- * Datentypen
    Term (..),
    Op (..),
    -- * Beispiele
    term1,
    term2,
    term3
) where

-- | Ein Datentyp fuer einfache Ausdruecke.
data Term
    = Const  Integer       -- ^ Ein konstanter Ausdruck
    | Binary Op Term Term  -- ^ Zwei Ausdruecke, die mit einem binaeren
                           --   Operator verknuepft sind.
    deriving Show

-- | Ein Datentyp fuer binaere Operatoren.
data Op
    = Add -- ^ Addition
    | Sub -- ^ Subtraktion
    | Div -- ^ Division
    | Mul -- ^ Multiplikation
    deriving Show

-- | Ein Beispielterm: 7
term1 :: Term
term1 = Const 7

-- | Ein Beispielterm: 12 + 5 * 6
term2 :: Term
term2 = Binary Add (Const 12) (Binary Mul (Const 5) (Const 6))

-- | Ein Beispielterm mit Division durch 0: 12 / 0
term3 :: Term
term3 = Binary Div (Const 1) (Const 0)

