-- Fehlerbehandlung
module Eval2a where

-- Die folgende Zeile versteckt (>>=).
import Prelude hiding ((>>=))

import Term

-- Ergebnistyp: Das Ergebnis der Berechnung, oder ein Fehler
data MayFail a = Result a | Error Cause
    deriving Show

-- Division durch Null ist die einzige Fehlerursache hier.
data Cause = DivisionByZero
    deriving Show

-- Idee:
(>>=) :: MayFail a -> (a -> MayFail b) -> MayFail b
(Error c)  >>= f = Error c
(Result r) >>= f = f r

-- | 'eval' berechnet den Wert eines Ausdrucks.
eval :: Term -> MayFail Integer
eval (Const a) = Result a

-- XXX: Der folgende Fall laesst sich mit (>>=) vereinfachen:
eval (Binary op a b) = case (eval a) of
    Error c -> Error c
    Result a' -> case (eval b) of
        Error c -> Error c
        Result b' -> evalOp op a' b'

-- | 'evalOp' berechnet das Ergebnis einer binaeren Operation,
-- angewandt auf zwei Operanden.
evalOp :: Op -> Integer -> Integer -> MayFail Integer
evalOp Add a b = Result (a + b)
evalOp Sub a b = Result (a - b)
evalOp Mul a b = Result (a * b)
evalOp Div a 0 = Error DivisionByZero
evalOp Div a b = Result (a `div` b)
