-- Zaehlen von Operationen
module Eval4b where

-- verstecke (>>=) und return.
import Term

-- 'CountOps a' enthaelt eine Funktion, die einen Anfangswert des Zaehlers
-- uebergeben bekommt, und das Ergebnis der Berechnung sowie den Endwert
-- des Zaehlers zurueckgibt.
data CountOps a = CountOps (Int -> (a, Int))

-- 'runCountOps' extrahiert die in 'CountOps a' enthaltene Funktion
runCountOps :: CountOps a -> Int -> (a, Int)
runCountOps (CountOps f) = f

instance Monad CountOps where
    CountOps x >>= f = CountOps (\cnt ->
        let (x', cnt') = x cnt
        in  runCountOps (f x') cnt')

    return a = CountOps (\cnt -> (a, cnt))

-- | 'step' zaehlt einen Rechenschritt
step :: CountOps ()
step = CountOps (\cnt -> ((), cnt + 1))

-- | 'eval' berechnet den Wert eines Ausdrucks.
eval :: Term -> CountOps Integer
eval (Const a) = return a
eval (Binary op a b) = do
    a' <- eval a
    b' <- eval b
    step
    evalOp op a' b'

-- | 'evalOp' berechnet das Ergebnis einer binaeren Operation,
-- angewandt auf zwei Operanden.
evalOp :: Op -> Integer -> Integer -> CountOps Integer
evalOp Add a b = return (a + b)
evalOp Sub a b = return (a - b)
evalOp Mul a b = return (a * b)
evalOp Div a b = return (a `div` b)
