-- Aufzeichnung von Zwischenschritten
module Eval6 where

import Term

import Control.Monad.Writer

type Log a = WriterT [LogEntry] Maybe a
type LogEntry = String

data Result a = None | Result [LogEntry] a

instance Show a => Show (Result a) where
    show None = "Error result."
    show (Result log a) = unlines (log ++ ["-->", show a])

-- | 'addLog' fuegt einen Eintrag zum Protokoll hinzu
addLog :: LogEntry -> Log ()
addLog e = tell [e]

eval' :: Term -> Result Integer
eval' t = case runWriterT (eval t) of
    Nothing       -> None
    Just (a, log) -> Result log a

-- | 'eval' berechnet den Wert eines Ausdrucks.
eval :: Term -> Log Integer
eval (Const a) = return a
eval (Binary op a b) = do
    a' <- eval a
    b' <- eval b
    logOp op a' b'
    evalOp op a' b'

-- | 'evalOp' berechnet das Ergebnis einer binaeren Operation,
-- angewandt auf zwei Operanden.
evalOp :: Op -> Integer -> Integer -> Log Integer
evalOp Add a b = return (a + b)
evalOp Sub a b = return (a - b)
evalOp Mul a b = return (a * b)
evalOp Div a 0 = mzero
evalOp Div a b = return (a `div` b)

-- | 'logOp' erzeugt einen Logeintrag fuer eine binaere Operation
logOp :: Op -> Integer -> Integer -> Log ()
logOp op a b = addLog (unwords [show op, show a, show b])
