-- Zaehlen von Operationen
module Eval3a where

-- Verstecke (>>=) und return.
import Prelude hiding ((>>=), return)

import Term

-- 'CountOps a' ist eine Funktion, die einen Anfangswert des Zaehlers
-- uebergeben bekommt, und das Ergebnis der Berechnung sowie den Endwert
-- des Zaehlers zurueckgibt.
type CountOps a = Int -> (a, Int)

-- Idee:
(>>=) :: CountOps a -> (a -> CountOps b) -> CountOps b
x >>= f = \cnt ->
    let (x', cnt') = x cnt
    in  (f x') cnt'

return :: a -> CountOps a
return a = \cnt -> (a, cnt)

-- | 'step' zaehlt einen Rechenschritt
step :: CountOps ()
step = \cnt -> ((), cnt + 1)

-- | 'eval' berechnet den Wert eines Ausdrucks.
eval :: Term -> CountOps Integer
eval (Const a) = return a
eval (Binary op a b) =
    eval a >>= \a' ->
    eval b >>= \b' ->
    step   >>= \_ ->
    evalOp op a' b'

-- | 'evalOp' berechnet das Ergebnis einer binaeren Operation,
-- angewandt auf zwei Operanden.
evalOp :: Op -> Integer -> Integer -> CountOps Integer
evalOp Add a b = return (a + b)
evalOp Sub a b = return (a - b)
evalOp Mul a b = return (a * b)
evalOp Div a b = return (a `div` b)
