-- Fehlerbehandlung
module Eval4a where

import Term

-- Ergebnistyp: Das Ergebnis der Berechnung, oder ein Fehler
data MayFail a = Result a | Error Cause
    deriving Show

-- Division durch Null ist die einzige Fehlerursache hier.
data Cause = DivisionByZero
    deriving Show

instance Monad MayFail where
    (Error c)  >>= f = Error c
    (Result r) >>= f = f r

    return r = Result r

-- | 'eval' berechnet den Wert eines Ausdrucks.
eval :: Term -> MayFail Integer
eval (Const a) = return a
eval (Binary op a b) = do
    a' <- eval a
    b' <- eval b
    evalOp op a' b'

-- | 'evalOp' berechnet das Ergebnis einer binaeren Operation,
-- angewandt auf zwei Operanden.
evalOp :: Op -> Integer -> Integer -> MayFail Integer
evalOp Add a b = return (a + b)
evalOp Sub a b = return (a - b)
evalOp Mul a b = return (a * b)
evalOp Div a 0 = Error DivisionByZero
evalOp Div a b = return (a `div` b)
