module TermParser (parseTerm) where

import Parser
import Term

token :: Parser a -> Parser a
token p = many space >> p

-- parst einen kompletten Ausdruck
parseTerm :: Parser Term
parseTerm = do
    t <- parseAdd
    token eof
    return t

-- parst einen additiven Ausdruck
parseAdd :: Parser Term
parseAdd = parseMul >>= parseAdd'

parseAdd' :: Term -> Parser Term
parseAdd' t1 = return t1
  <|> do
    op <- token (satisfy (`elem` "+-"))
    t2 <- parseMul
    parseAdd' (Binary (if op == '+' then Add else Sub) t1 t2)

-- parst einen multiplikativen Ausdruck
parseMul :: Parser Term
parseMul = parseNum >>= parseMul'

parseMul' :: Term -> Parser Term
parseMul' t1 = return t1
  <|> do
    op <- token (satisfy (`elem` "*/"))
    t2 <- parseNum
    parseMul' (Binary (if op == '*' then Mul else Div) t1 t2)

-- parst eine Zahl oder einen geklammerten Ausdruck
parseNum :: Parser Term
parseNum = fmap (Const . read) (token (many1 digit))
  <|> do
     token (char '(')
     t <- parseAdd
     token (char ')')
     return t
