module Computation (Computation,
		    halt, Computation.return,
		    liftStateStack, liftStateHeap,
		    transferControl)
    where

import Memory (Univ(..), Memory(..))
import Heap (Heap(..))
import Stack (Stack)

import Control.Monad.State


data Progress a = Halted
		  | Returned Memory
		  | Running a

instance Functor Progress where
    fmap _ Halted = Halted
    fmap _ (Returned x) = Returned x
    fmap f (Running x) = Running (f x)
instance Monad Progress where
    Halted     >>= _ = Halted
    Returned x >>= _ = Returned x
    Running x  >>= f = f x
    return x         = Running x
    fail s           = Halted

type Computation = StateT Memory Progress

halt, return :: Computation a
halt = StateT (const Halted)
return = StateT Returned

liftStateStack :: State Stack a -> Computation a
liftStateStack op = StateT doIt
    where doIt mem = let (y, s) = runState op (memStack mem)
		       in Running (y, mem { memStack = s })

liftStateHeap :: State (Heap Univ) a -> Computation a
liftStateHeap op = StateT doIt
    where doIt mem = let (y, h) = runState op (memHeap mem)
			 in Running (y, mem { memHeap = h })

transferControl :: Computation () -> Computation ()
transferControl = mapStateT cont
    where cont Halted = Halted
	  cont (Returned x) = Running ((), x)
	  cont (Running ((), x)) = Running ((), x)
