module Equiv where

newtype Equiv a = Equiv { equivEQ :: a -> a -> Bool }


initialEquiv :: Eq a => Equiv a
initialEquiv = Equiv (==)

finalEquiv :: Equiv a
finalEquiv = Equiv (const (const True))

kernelEquiv :: (a -> b) -> Equiv b -> Equiv a
kernelEquiv f (Equiv eq) = Equiv eq'
    where eq' x y = eq (f x) (f y)

andEquiv :: Equiv a -> Equiv a -> Equiv a
andEquiv (Equiv eq1) (Equiv eq2) = Equiv eq
    where eq x y = eq1 x y && eq2 x y

liftMaybeEquiv :: Equiv a -> Equiv (Maybe a)
liftMaybeEquiv (Equiv eq) = Equiv eq'
    where eq' Nothing Nothing = True
	  eq' (Just x1) (Just x2) = eq x1 x2
	  eq' _ _ = False

liftListEquiv :: Equiv a -> Equiv [a]
liftListEquiv (Equiv eq) = Equiv eq'
    where eq' [] [] = True
	  eq' (x1 : ys1) (x2 : ys2) = eq x1 x2 && eq' ys1 ys2
	  eq' _ _ = False
