module Memory where

import Signature (Signature(..))
import Heap      (Heap, emptyHeap)
import Stack     (Stack, emptyStack)
import Value     (Value)
import Program   (Function, ConstantId)

import Control.Monad (zipWithM_)
import Data.FiniteMap


instance (Show key, Show elt) => Show (FiniteMap key elt) where
  show m = show (fmToList m)

data Univ a = UTuple (FiniteMap String a)
	    | UCotuple String (Maybe a)
	    | UClosure Function [a]
            deriving Show

instance Functor Univ where
    fmap f (UTuple m) = UTuple (mapFM (const f) m)
    fmap f (UCotuple l b) = UCotuple l (fmap f b)
    fmap f (UClosure g xs) = UClosure g (fmap f xs)

instance Signature Univ where
    UTuple m1 ==* UTuple m2 = m1 == m2
    UCotuple l1 b1 ==* UCotuple l2 b2 = l1 == l2 && b1 == b2
    UClosure f1 xs1 ==* UClosure f2 xs2 = f1 == f2 && xs1 == xs2
    _ ==* _ = False
    fmapM f (UTuple m) =
	mapM f' (fmToList m) >>= return . UTuple . listToFM
	where f' (k, e) = do e' <- f e
			     return (k, e')
    fmapM f (UCotuple l Nothing) = return (UCotuple l Nothing)
    fmapM f (UCotuple l (Just b)) = f b >>= return . UCotuple l . Just
    fmapM f (UClosure g xs) = mapM f xs >>= return . UClosure g
    ffold f e (UTuple m) = foldFM (const (flip f)) e m
    ffold f e (UCotuple l Nothing) = e
    ffold f e (UCotuple l (Just b)) = f e b
    ffold f e (UClosure g xs) = foldl f e xs
    fzipWithM_ f (UTuple m1) (UTuple m2) = zipWithM_ f (eltsFM m1) (eltsFM m2)
    fzipWithM_ f (UCotuple l1 Nothing) (UCotuple l2 Nothing)
	| l1 == l2 = return ()
    fzipWithM_ f (UCotuple l1 (Just b1)) (UCotuple l2 (Just b2))
	| l1 == l2 = f b1 b2
    fzipWithM_ f (UClosure g1 xs1) (UClosure g2 xs2)
	| g1 == g2 = zipWithM_ f xs1 xs2

data Memory = Memory
    { memHeap :: Heap Univ,
      memStack :: Stack,
      memConstants :: FiniteMap ConstantId Value
    }

emptyMemory = Memory { memHeap = emptyHeap,
		       memStack = emptyStack,
		       memConstants = emptyFM }

