module NatList where

import Signature
import Dialgebra


type NatSig = Const () :+: Identity

data NatCarrier = Zero | Succ NatCarrier
     deriving (Show, Eq)

instance Dialgebra NatSig NatCarrier where
  cons (Inject1 (Const ()))   = Zero
  cons (Inject2 (Identity n)) = Succ n
  decons Zero     = Inject1 (Const ())
  decons (Succ n) = Inject2 (Identity n)


type ListSig a = Const () :+: (Const a :*: Identity)

mapterm :: (a -> b) -> ListSig a c -> ListSig b c
mapterm f (Inject1 (Const ())) = Inject1 (Const ())
mapterm f (Inject2 (Const x :*: Identity l)) =
    Inject2 (Const (f x) :*: Identity l)

data ListCarrier a = Nil | Cons a (ListCarrier a)
     deriving (Show, Eq)

instance Eq a => Dialgebra (ListSig a) (ListCarrier a) where
  cons (Inject1 (Const ()))               = Nil
  cons (Inject2 (Const x :*: Identity l)) = Cons x l
  decons Nil        = Inject1 (Const ())
  decons (Cons x l) = Inject2 (Const x :*: Identity l)


type UListSig = Const () :+: (Identity :*: Identity)

data UListCarrier = UNil | UCons UListCarrier UListCarrier
     deriving (Show, Eq)

instance Dialgebra UListSig UListCarrier where
  cons (Inject1 (Const ()))                  = UNil
  cons (Inject2 (Identity x :*: Identity l)) = UCons x l
  decons UNil        = Inject1 (Const ())
  decons (UCons x l) = Inject2 (Identity x :*: Identity l)


type NatListSig = NatSig :+: UListSig

data NatListCarrier = XZero | XNil
		    | XSucc NatListCarrier
		    | XCons NatListCarrier NatListCarrier
     deriving (Show, Eq)

instance Dialgebra NatListSig NatListCarrier where
  cons (Inject1 (Inject1 (Const ())))                  = XZero
  cons (Inject1 (Inject2 (Identity n)))                = XSucc n
  cons (Inject2 (Inject1 (Const ())))                  = XNil
  cons (Inject2 (Inject2 (Identity x :*: Identity l))) = XCons x l
  decons XZero       = Inject1 (Inject1 (Const ()))
  decons (XSucc n)   = Inject1 (Inject2 (Identity n))
  decons XNil        = Inject2 (Inject1 (Const ()))
  decons (XCons x l) = Inject2 (Inject2 (Identity x :*: Identity l))


liftNat :: Decons NatListSig NatCarrier
liftNat Zero     = Inject1 (Inject1 (Const ()))
liftNat (Succ n) = Inject1 (Inject2 (Identity n))

liftNatAna :: NatCarrier -> NatListCarrier
liftNatAna = ana liftNat


type NatOrList = Either NatCarrier (ListCarrier NatCarrier)

lift :: Decons NatListSig NatOrList
lift (Left Zero)        = Inject1 (Inject1 (Const ()))
lift (Left (Succ n))    = Inject1 (Inject2 (Identity (Left n)))
lift (Right Nil)        = Inject2 (Inject1 (Const ()))
lift (Right (Cons x l)) = Inject2 (Inject2 (Identity (Left x) :*:
					    Identity (Right l)))

liftAna :: NatOrList -> NatListCarrier
liftAna = ana lift

liftNatAna' = liftAna . Left
liftListAna = liftAna . Right


asNat :: Decons NatSig NatListCarrier
asNat XZero = Inject1 (Const ())
asNat (XSucc n) = Inject2 (Identity n)

asNatAna :: NatListCarrier -> NatCarrier
asNatAna = ana asNat


asList :: Decons (ListSig NatCarrier) NatListCarrier
asList XNil        = Inject1 (Const ())
asList (XCons x l) = Inject2 (Const (asNatAna x) :*: Identity l)

asListAna :: NatListCarrier -> ListCarrier NatCarrier
asListAna = ana asList


unlift :: Decons (NatSig :+: ListSig NatListCarrier) NatListCarrier
unlift XZero       = Inject1 (Inject1 (Const ()))
unlift (XSucc n)   = Inject1 (Inject2 (Identity n))
unlift XNil        = Inject2 (Inject1 (Const ()))
unlift (XCons r l) = Inject2 (Inject2 (Const r :*: Identity l))


asNatAna' :: NatListCarrier -> NatCarrier
asNatAna' = ana (part1 . unlift)

asListAna' :: NatListCarrier -> ListCarrier NatCarrier
asListAna' = ana (mapterm asNatAna' . part2 . unlift)


mylength :: Decons NatSig NatListCarrier
mylength XNil = Inject1 (Const ())
mylength (XCons x l) = Inject2 (Identity l)

lengthAna :: NatListCarrier -> NatCarrier
lengthAna = ana mylength

mylength' :: Decons NatListSig NatListCarrier
mylength' = Inject1 . mylength

lengthAna' :: NatListCarrier -> NatListCarrier
lengthAna' = ana mylength'

lengthAna'' :: NatListCarrier -> NatCarrier
lengthAna'' = asNatAna . lengthAna'
